//===-- Opcodes.cpp - The Cheerp JavaScript generator ---------------------===//
//
//                     Cheerp: The C++ compiler for the Web
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
// Copyright 2014 Leaning Technologies
//
//===----------------------------------------------------------------------===//

#include "llvm/ADT/StringExtras.h"
#include "llvm/Cheerp/NameGenerator.h"
#include "llvm/Cheerp/GlobalDepsAnalyzer.h"
#include "llvm/Cheerp/Utility.h"
#include "llvm/IR/Function.h"
#include <functional>
#include <set>

using namespace llvm;

namespace cheerp {

/**
 * Note: this describes the valid *C++* identifiers.
 * 
 * Iterate over all the valid JS identifier is much more complicated , because JS uses unicode.
 * Reference for valid JS identifiers:
*  http://stackoverflow.com/questions/1661197/valid-characters-for-javascript-variable-names
*/
struct JSSymbols
{
	enum : char {first_symbol = 'a' };
	static char next( char c )
	{
		return  c == '$' ? 'a' :
			c == 'z' ? 'A' :
			c == 'Z' ? '0' :
			c == '9' ? '_' :
			c == '_' ? '$' :
			++c;
	}

	JSSymbols(const std::vector<std::string>& reservedNames):externallyReservedNames(reservedNames)
	{
	}

	template< class String >
	bool is_valid( String & s )
	{
		// Can not be empty
		if ( s.empty() ) return false;
		
		// Can not start with a digit
		if ( std::isdigit(s.front() ) )
		{
			std::fill( s.begin(), s.end(), '_' );
			return false;
		}
		
		// Can not be made only of '_';
		if ( std::all_of( s.begin(), s.end(), [](char c) { return c == '_'; }) )
			return false;
		
		// Check for reserved keywords
		if ( is_reserved_name(s) )
			return false;
		
		// "null" is used by cheerp internally
		if ( s == "null" )
			return false;
		
		// In the rare case that a name is longer than 4 characters, it need to start with an underscore
		// just to be safe.
		if ( s.size() > 4 && s.front() != '_' )
		{
			std::fill( s.begin(), s.end(), '_' );
			return false;
		}
		
		// Check for labels generated by the relooper
		if ( s.front() == 'L' && std::all_of( s.begin()+1, s.end(), ::isdigit ) )
		{
			std::fill( s.begin()+1,s.end(),'9');
			return false;
		}

		return true;
	}

	template< class String >
	bool is_reserved_name( String& s)
	{
		const char* reserved_names[] = {
			"byte",
			"case",
			"char",
			"do",
			"else",
			"enum",
			"for",
			"goto",
			"if",
			"in",
			"int",
			"let",
			"new",
			"this",
			"top",
			"try",
			"var",
			"void",
			"with"
		};
		return std::binary_search(reserved_names, reserved_names+(sizeof(reserved_names)/sizeof(const char*)), s) ||
			std::binary_search(externallyReservedNames.begin(), externallyReservedNames.end(), s);
	}
	const std::vector<std::string>& externallyReservedNames;
};

NameGenerator::NameGenerator(const Module& M, const GlobalDepsAnalyzer& gda, const Registerize& r,
				const PointerAnalyzer& PA, const std::vector<std::string>& reservedNames, bool makeReadableNames):registerize(r), PA(PA), reservedNames(reservedNames)
{
	if ( makeReadableNames )
		generateReadableNames(M, gda);
	else
		generateCompressedNames(M, gda);
	generateTypeNames(gda);
}

llvm::StringRef NameGenerator::getNameForEdge(const llvm::Value* v) const
{
	assert(!edgeContext.isNull());
	if (const Instruction* I=dyn_cast<Instruction>(v))
	{
		auto it=edgeNamemap.find(InstOnEdge(edgeContext.fromBB, edgeContext.toBB, registerize.getRegisterId(I)));
		if (it!=edgeNamemap.end())
			return it->second;
	}
	return namemap.at(v);
}

llvm::StringRef NameGenerator::getSecondaryNameForEdge(const llvm::Value* v) const
{
	assert(!edgeContext.isNull());
	if (const Instruction* I=dyn_cast<Instruction>(v))
	{
		auto it=edgeSecondaryNamemap.find(InstOnEdge(edgeContext.fromBB, edgeContext.toBB, registerize.getRegisterId(I)));
		if (it!=edgeSecondaryNamemap.end())
			return it->second;
	}
	return secondaryNamemap.at(v);
}

SmallString< 4 > NameGenerator::filterLLVMName(StringRef s, NAME_FILTER_MODE filterMode)
{
	SmallString< 4 > ans;
	ans.reserve( s.size() + 1 );

	//Add an '_', 'L' or 'M' to skip reserved names
	switch(filterMode)
	{
		case GLOBAL:
			ans.push_back( '_' );
			break;
		case GLOBAL_SECONDARY:
			ans.push_back( '$' );
			break;
		case LOCAL:
			ans.push_back( 'L' );
			break;
		case LOCAL_SECONDARY:
			ans.push_back( 'M' );
			break;
	}

	for ( char c : s )
	{
		//We need to escape invalid chars
		switch(c)
		{
			case '.':
				ans.append("$p");
				break;
			case '-':
				ans.append("$m");
				break;
			case ':':
				ans.append("$c");
				break;
			case '<':
				ans.append("$l");
				break;
			case '>':
				ans.append("$r");
				break;
			case ' ':
				ans.append("$s");
				break;
			default:
				ans.push_back(c);
		}
	}

	return ans;
}

void NameGenerator::generateCompressedNames(const Module& M, const GlobalDepsAnalyzer& gda)
{
	typedef std::pair<unsigned, const Value *> useValuePair;
	typedef std::pair<unsigned, std::vector<const Value *> > useValuesPair;
	typedef std::vector<useValuesPair> useValuesVec;
	typedef std::pair<unsigned, std::vector<InstOnEdge> > useInstsOnEdgePair;
	typedef std::vector<useInstsOnEdgePair> useInstsOnEdgeVec;
        
	// Class to handle giving names to temporary variables needed for recursively dependent PHIs
	class CompressedPHIHandler: public EndOfBlockPHIHandler
	{
	public:
		CompressedPHIHandler(const BasicBlock* f, const BasicBlock* t, NameGenerator& n, useInstsOnEdgeVec& a ):
			EndOfBlockPHIHandler(n.PA), fromBB(f), toBB(t), namegen(n), allTmpPHIs(a), nextIndex(0)
		{
		}
	private:
		const BasicBlock* fromBB;
		const BasicBlock* toBB;
		NameGenerator& namegen;
		useInstsOnEdgeVec& allTmpPHIs;
		uint32_t nextIndex;
		void handleRecursivePHIDependency(const Instruction* incoming) override
		{
			uint32_t regId=namegen.registerize.getRegisterId(incoming);
			// We don't know exactly how many times the tmpphi is going to be used in this edge
			// but assume 1. We increment the usage count for the first not already used tmpphi
			// and add the InstOnEdge to its list
			if (nextIndex >= allTmpPHIs.size())
				allTmpPHIs.resize(nextIndex+1);
			allTmpPHIs[nextIndex].first++;
			allTmpPHIs[nextIndex].second.emplace_back(InstOnEdge(fromBB, toBB, regId));
			nextIndex++;
		}
		void handlePHI(const Instruction* phi, const Value* incoming) override
		{
			// Nothing to do here, we have already given names to all PHIs
		}
	};
	/**
	 * Collect the local values.
	 * 
	 * We sort them by uses, then store together those in the same position.
	 * i.e. allLocalValues[0].second will contain all the most used local values
	 * for each function, and allLocalValues[0].first will be the sum of the uses
	 * of all those local values.
	 */
        
	useValuesVec allLocalValues;
	useInstsOnEdgeVec allTmpPHIs;
        
	/**
	 * Sort the global values by number of uses
	 */
	std::set< useValuePair, std::greater< useValuePair > > allGlobalValues;

	for (const Function & f : M.getFunctionList() )
	{
		unsigned nUses = f.getNumUses();

		if ( f.getName() == "_Z7webMainv" )
			++nUses; // We explicitly invoke the webmain

		// Constructors are also explicitly invoked
		if ( std::find(gda.constructors().begin(), gda.constructors().end(), &f ) != gda.constructors().end() )
			++nUses;

		allGlobalValues.emplace( nUses, &f );

		/**
		 * TODO, some cheerp-internals functions are actually generated even with an empty IR.
		 * They should be considered here.
		 */
		if ( f.empty() ) 
			continue;

		// Local values are all stored in registers
		useValuesVec thisFunctionLocals;

		// Insert all the instructions
		for (const BasicBlock & bb : f)
		{
			for (const Instruction & I : bb)
			{
				if ( needsName(I, PA) )
				{
					uint32_t registerId = registerize.getRegisterId(&I);
					if (registerId >= thisFunctionLocals.size())
						thisFunctionLocals.resize(registerId+1);
					useValuesPair& regData = thisFunctionLocals[registerId];
					// Add the uses for this instruction to the total count for the register
					regData.first+=I.getNumUses();
					// Add the instruction itself to the list of istructions
					regData.second.push_back(&I);
				}
			}
			// Handle the special names required for the edges between blocks
			const TerminatorInst* term=bb.getTerminator();
			for(uint32_t i=0;i<term->getNumSuccessors();i++)
			{
				const BasicBlock* succBB=term->getSuccessor(i);
				CompressedPHIHandler(&bb, succBB, *this, allTmpPHIs).runOnEdge(registerize, &bb, succBB);
			}
		}

		uint32_t currentArgPos=thisFunctionLocals.size();
		thisFunctionLocals.resize(currentArgPos+f.arg_size());
		// Insert the arguments
		for ( auto arg_it = f.arg_begin(); arg_it != f.arg_end(); ++arg_it, currentArgPos++ )
		{
			thisFunctionLocals[currentArgPos].first = f.getNumUses();
			thisFunctionLocals[currentArgPos].second.push_back( arg_it );
		}

		// Resize allLocalValues so that we have empty useValuesPair at the end of the container
		if ( thisFunctionLocals.size() > allLocalValues.size() )
			allLocalValues.resize( thisFunctionLocals.size() );

		std::sort(thisFunctionLocals.begin(),thisFunctionLocals.end());
		auto dst_it = allLocalValues.begin();

		for (auto src_it = thisFunctionLocals.rbegin(); src_it != thisFunctionLocals.rend(); ++src_it, ++dst_it )
		{
			dst_it->first += src_it->first;
			dst_it->second.insert(dst_it->second.end(), src_it->second.begin(), src_it->second.end());
		}
	}
        
	assert( std::is_sorted( 
		allLocalValues.rbegin(), 
		allLocalValues.rend(),
		[] (const useValuesPair & lhs, const useValuesPair & rhs) { return lhs.first < rhs.first; } 
		) );

	for ( const GlobalValue & GV : M.getGlobalList() )
	{
		if ( isa<GlobalVariable>(GV) && TypeSupport::isClientGlobal(&GV) )
		{
			demangler_iterator dmg( GV.getName() );
			assert(*dmg == "client");
			
			namemap.emplace( &GV, *(++dmg) );
			
			continue;
		}

		allGlobalValues.emplace( GV.getNumUses(), &GV );
	}

	/**
	 * Now generate the names and fill the namemap.
	 * 
	 * Note that there will never be a global with the same name as a local.
	 * This is suboptimal, since in theory we could check out for the uses
	 * of the global inside the function.
	 */
	name_iterator<JSSymbols> name_it((JSSymbols(reservedNames)));
	
	// We need to iterate over allGlobalValues, allLocalValues and allTmpPHIs
	// at the same time incrementing selectively only one of the iterators
	
	std::set< useValuePair >::const_iterator global_it = allGlobalValues.begin();
	useValuesVec::const_iterator local_it = allLocalValues.begin();
	useInstsOnEdgeVec::const_iterator tmpphi_it = allTmpPHIs.begin();

	bool globalsFinished = global_it == allGlobalValues.end();
	bool localsFinished = local_it == allLocalValues.end();
	bool tmpPHIsFinished = tmpphi_it == allTmpPHIs.end();
	for ( ; !globalsFinished || !localsFinished || !tmpPHIsFinished; ++name_it )
	{
		if ( !globalsFinished &&
			(localsFinished || global_it->first >= local_it->first) &&
			(tmpPHIsFinished || global_it->first >= tmpphi_it->first))
		{
			// Assign this name to a global value
			namemap.emplace( global_it->second, *name_it );
			// We need to consume another name to assign the secondary one
			if(needsSecondaryName(global_it->second, PA))
			{
				++name_it;
				secondaryNamemap.emplace( global_it->second, *name_it );
			}
			++global_it;
		}
		else if ( !localsFinished &&
			(globalsFinished || local_it->first >= global_it->first) &&
			(tmpPHIsFinished || local_it->first >= tmpphi_it->first))
		{
			// Assign this name to all the local values
			SmallString<4> primaryName = *name_it;
			SmallString<4> secondaryName;
			for ( const Value * v : local_it->second )
			{
				namemap.emplace( v, primaryName );
				// We need to consume another name to assign the secondary one
				if(needsSecondaryName(v, PA))
				{
					if(secondaryName.empty())
					{
						++name_it;
						secondaryName = *name_it;
					}
					secondaryNamemap.emplace( v, secondaryName );
				}
			}
			
			++local_it;
		}
		else
		{
			SmallString<4> primaryName = *name_it;
			++name_it;
			SmallString<4> secondaryName = *name_it;
			// Assign this name to all the tmpphis
			for ( const InstOnEdge& i : tmpphi_it->second )
			{
				edgeNamemap.emplace( i, StringRef(primaryName));
				// TODO: Only add the secondary name if required
				// TODO: Tmpphis should be properly managed in Registerize
				edgeSecondaryNamemap.emplace( i, StringRef(secondaryName));
			}
			
			++tmpphi_it;
		}
		globalsFinished = global_it == allGlobalValues.end();
		localsFinished = local_it == allLocalValues.end();
		tmpPHIsFinished = tmpphi_it == allTmpPHIs.end();
	}
}

void NameGenerator::generateReadableNames(const Module& M, const GlobalDepsAnalyzer& gda)
{
	// Class to handle giving names to temporary variables needed for recursively dependent PHIs
	class ReadablePHIHandler: public EndOfBlockPHIHandler
	{
	public:
		ReadablePHIHandler(const BasicBlock* f, const BasicBlock* t, NameGenerator& n ):
			EndOfBlockPHIHandler(n.PA), fromBB(f), toBB(t), namegen(n), nextIndex(0)
		{
		}
	private:
		const BasicBlock* fromBB;
		const BasicBlock* toBB;
		NameGenerator& namegen;
		uint32_t nextIndex;
		void handleRecursivePHIDependency(const Instruction* incoming) override
		{
			uint32_t regId=namegen.registerize.getRegisterId(incoming);
			namegen.edgeNamemap.emplace(InstOnEdge(fromBB, toBB, regId),
							StringRef( "tmpphi" + std::to_string(nextIndex)));
			// TODO: Only use secondary when necessary
			namegen.edgeSecondaryNamemap.emplace(InstOnEdge(fromBB, toBB, regId),
							StringRef( "tmpphi" + std::to_string(nextIndex++) + "o"));
		}
		void handlePHI(const Instruction* phi, const Value* incoming) override
		{
			// Nothing to do here, we have already given names to all PHIs
		}
	};
	for (const Function & f : M.getFunctionList() )
	{
		// Temporary mapping between registers and names
		// NOTE: We only store references to names in the namemap
		std::unordered_map<uint32_t, llvm::SmallString<4>& > regmap;

		for (const BasicBlock & bb : f)
		{
			for (const Instruction & I : bb)
			{
				if ( needsName(I, PA) )
				{
					uint32_t registerId = registerize.getRegisterId(&I);
					// If the register already has a name, use it
					// Otherwise assign one as good as possible and assign it to the register as well
					auto regNameIt = regmap.find(registerId);
					StringRef primaryName;
					if(regNameIt != regmap.end())
					{
						primaryName = namemap.emplace( &I, regNameIt->second ).first->second;
					}
					else if ( I.hasName() )
					{
						auto& name = namemap.emplace( &I, filterLLVMName(I.getName(), LOCAL) ).first->second;
						regmap.emplace( registerId, name );
						primaryName = name;
					}
					else
					{
						auto& name = namemap.emplace( &I, StringRef( "tmp" + std::to_string(registerId) ) ).first->second;
						regmap.emplace( registerId, name );
						primaryName = name;
					}
					if(needsSecondaryName(&I, PA))
						secondaryNamemap.emplace( &I, StringRef((primaryName+"o").str()));
				}
			}
			// Handle the special names required for the edges between blocks
			const TerminatorInst* term=bb.getTerminator();
			for(uint32_t i=0;i<term->getNumSuccessors();i++)
			{
				const BasicBlock* succBB=term->getSuccessor(i);
				ReadablePHIHandler(&bb, succBB, *this).runOnEdge(registerize, &bb, succBB);
			}
		}

		for ( auto arg_it = f.arg_begin(); arg_it != f.arg_end(); ++arg_it )
		{
			bool needsTwoNames = needsSecondaryName(arg_it, PA);
			if ( arg_it->hasName() )
			{
				namemap.emplace( arg_it, filterLLVMName(arg_it->getName(), LOCAL) );
				if(needsTwoNames)
					secondaryNamemap.emplace( arg_it, filterLLVMName(arg_it->getName(), LOCAL_SECONDARY) );
			}
			else
			{
				namemap.emplace( arg_it, StringRef( "Larg" + std::to_string(arg_it->getArgNo()) ) );
				if(needsTwoNames)
					secondaryNamemap.emplace( arg_it, StringRef( "Marg" + std::to_string(arg_it->getArgNo()) ) );
			}
		}
			
		namemap.emplace( &f, filterLLVMName( f.getName(), GLOBAL ) );
	}

	for (const GlobalVariable & GV : M.getGlobalList() )
		if (TypeSupport::isClientGlobal(&GV) )
		{
			demangler_iterator dmg( GV.getName() );
			assert(*dmg == "client");
			
			namemap.emplace( &GV, *(++dmg) );
			
		}
		else
		{
			namemap.emplace( &GV, filterLLVMName( GV.getName(), GLOBAL ) );
			bool needsTwoNames = needsSecondaryName(&GV, PA);
			if(needsTwoNames)
				secondaryNamemap.emplace( &GV, filterLLVMName(GV.getName(), GLOBAL_SECONDARY) );
		}
}

void NameGenerator::generateTypeNames(const GlobalDepsAnalyzer& gda)
{
	for(Type* T: gda.classesUsed())
	{
		if(isa<StructType>(T) && cast<StructType>(T)->hasName())
			typemap.insert(std::make_pair(T, filterLLVMName(cast<StructType>(T)->getName(), GLOBAL)));
		else
			typemap.insert(std::make_pair(T, StringRef("_literal" + std::to_string(typemap.size()))));
	}
	for(Type* T: gda.dynAllocArrays())
	{
		if(isa<StructType>(T) && cast<StructType>(T)->hasName())
			typemap.insert(std::make_pair(T, filterLLVMName(cast<StructType>(T)->getName(), GLOBAL)));
		else
			typemap.insert(std::make_pair(T, StringRef("_literal" + std::to_string(typemap.size()))));
	}
}

bool NameGenerator::needsName(const Instruction & I, const PointerAnalyzer& PA) const
{
	return !isInlineable(I, PA) && !I.getType()->isVoidTy() && !I.use_empty();
}

bool NameGenerator::needsSecondaryName(const Value* V, const PointerAnalyzer& PA) const
{
	if(!V->getType()->isPointerTy())
		return false;
	if(PA.getPointerKind(V) == SPLIT_REGULAR && !PA.getConstantOffsetForPointer(V))
		return true;
	return false;
}

}
